/*
** send.c -- send a composed message
**
** This code is Copyright (c) 2002, by the authors of nmh.  See the
** COPYRIGHT file in the root directory of the nmh distribution for
** complete copyright information.
*/

#include <h/mh.h>
#include <fcntl.h>
#include <errno.h>
#include <signal.h>
#include <h/signals.h>
#include <setjmp.h>
#include <h/mime.h>
#include <h/tws.h>
#include <h/utils.h>

#ifdef TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# ifdef TM_IN_SYS_TIME
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif

int debugsw = 0;  /* global */
int forwsw  = 1;
int pushsw  = 0;
int verbsw  = 0;
char *altmsg   = NULL;
char *annotext = NULL;
char *distfile = NULL;

static jmp_buf env;

/* name of temp file for body content */
static char body_file_name[MAXPATHLEN + 1];
/* name of mhbuild composition temporary file */
static char composition_file_name[MAXPATHLEN + 1]; 
static int field_size;  /* size of header field buffer */
static char *field;  /* header field buffer */
static FILE *draft_file;  /* draft file pointer */
static FILE *body_file;  /* body file pointer */
static FILE *composition_file;  /* composition file pointer */

/*
** static prototypes
*/
static int sendsbr(char **, int, char *, struct stat *, int);
static void armed_done(int) NORETURN;
static void alert(char *, int);
static int tmp_fd(void);
static void anno(struct stat *);
static int sendaux(char **, int, char *, struct stat *);
static int attach(char *);
static void clean_up_temporary_files(void);
static int get_line(void);
static void make_mime_composition_file_entry(char *);


static struct swit switches[] = {
#define ALIASW  0
	{ "alias aliasfile", 0 },
#define DEBUGSW  1
	{ "debug", -5 },
#define FORWSW  2
	{ "forward", 0 },
#define NFORWSW  3
	{ "noforward", 0 },
#define PUSHSW  4
	{ "push", 0 },
#define NPUSHSW  5
	{ "nopush", 0 },
#define VERBSW  6
	{ "verbose", 0 },
#define NVERBSW  7
	{ "noverbose", 0 },
#define WATCSW  8
	{ "watch", 0 },
#define NWATCSW  9
	{ "nowatch", 0 },
#define VERSIONSW  10
	{ "version", 0 },
#define HELPSW  11
	{ "help", 0 },
	{ NULL, 0 }
};


int
main(int argc, char **argv)
{
	int msgp = 0, nfiles = 0, distsw = 0, vecp = 1;
	int msgnum, status;
	int in, out;
	char *cp, *maildir = NULL;
	char buf[BUFSIZ], **ap, **argp, **arguments;
	char *msgs[MAXARGS], *vec[MAXARGS];
	char *files[MAXARGS];
	struct msgs *mp;
	struct stat st;
	struct stat st2;


#ifdef LOCALE
	setlocale(LC_ALL, "");
#endif
	invo_name = mhbasename(argv[0]);

	/* read user profile/context */
	context_read();

	arguments = getarguments(invo_name, argc, argv, 1);
	argp = arguments;

	vec[vecp++] = "-library";
	vec[vecp++] = getcpy(toabsdir("+"));

	while ((cp = *argp++)) {
		if (*cp == '-') {
			switch (smatch(++cp, switches)) {
			case AMBIGSW:
				ambigsw(cp, switches);
				done(1);
			case UNKWNSW:
				adios(NULL, "-%s unknown\n", cp);

			case HELPSW:
				snprintf(buf, sizeof(buf),
						"%s [file] [switches]",
						invo_name);
				print_help(buf, switches, 1);
				done(1);
			case VERSIONSW:
				print_version(invo_name);
				done(1);

			case PUSHSW:
				pushsw++;
				continue;
			case NPUSHSW:
				pushsw = 0;
				continue;

			case FORWSW:
				forwsw++;
				continue;
			case NFORWSW:
				forwsw = 0;
				continue;

			case VERBSW:
				verbsw++;
				vec[vecp++] = --cp;
				continue;
			case NVERBSW:
				verbsw = 0;
				vec[vecp++] = --cp;
				continue;

			case DEBUGSW:
				debugsw++;  /* fall */
			case WATCSW:
			case NWATCSW:
				vec[vecp++] = --cp;
				continue;

			case ALIASW:
				vec[vecp++] = --cp;
				if (!(cp = *argp++) || *cp == '-') {
					adios(NULL, "missing argument to %s",
							argp[-2]);
				}
				vec[vecp++] = cp;
				continue;

			}
		} else {
			if (*cp == '/') {
				files[nfiles++] = cp;
			} else {
				msgs[msgp++] = cp;
			}
		}
	}

	/* check for "Aliasfile:" profile entry */
	if ((cp = context_find("Aliasfile"))) {
		char *dp = NULL;

		for (ap=brkstring(dp=getcpy(cp), " ", "\n"); ap && *ap; ap++) {
			vec[vecp++] = "-alias";
			vec[vecp++] = getcpy(etcpath(*ap));
		}
	}

	if (!msgp && !nfiles) {
		msgs[msgp++] = seq_cur;
	}
	maildir = toabsdir(draftfolder);

	if (chdir(maildir) == NOTOK) {
		adios(maildir, "unable to change directory to");
	}

	if (!(mp = folder_read(draftfolder))) {
		adios(NULL, "unable to read draft folder %s", draftfolder);
	}
	if (mp->nummsg == 0) {
		adios(NULL, "no messages in draft folder %s", draftfolder);
	}
	/* parse all the message ranges/sequences and set SELECTED */
	for (msgnum = 0; msgnum < msgp; msgnum++) {
		if (!m_convert(mp, msgs[msgnum])) {
			done(1);
		}
	}
	seq_setprev(mp);

	for (msgp = 0, msgnum = mp->lowsel; msgnum <= mp->hghsel; msgnum++) {
		if (is_selected(mp, msgnum)) {
			files[nfiles++] = getcpy(m_name(msgnum));
			unset_exists(mp, msgnum);
		}
	}

	mp->msgflags |= SEQMOD;
	seq_save(mp);

	if (!(cp = getenv("SIGNATURE")) || !*cp) {
		if ((cp = context_find("signature")) && *cp) {
			m_putenv("SIGNATURE", cp);
		}
	}

	for (msgnum = 0; msgnum < nfiles; msgnum++) {
		if (stat(files[msgnum], &st) == NOTOK) {
			adios(files[msgnum], "unable to stat draft file");
		}
	}

	if (!(annotext = getenv("mhannotate")) || !*annotext) {
		annotext = NULL;
	}
	if (!(altmsg = getenv("mhaltmsg")) || !*altmsg) {
		altmsg = NULL;  /* used by dist interface - see below */
	}

	if ((cp = getenv("mhdist")) && *cp && (distsw = atoi(cp)) && altmsg) {
		vec[vecp++] = "-dist";
		if ((in = open(altmsg, O_RDONLY)) == NOTOK) {
			adios(altmsg, "unable to open for reading");
		}
		fstat(in, &st2);
		distfile = getcpy(m_mktemp2(NULL, invo_name, NULL, NULL));
		if ((out = creat(distfile, (int)st2.st_mode & 0777))==NOTOK) {
			adios(distfile, "unable to open for writing");
		}
		cpydata(in, out, altmsg, distfile);
		close(in);
		close(out);
	} else {
		distfile = NULL;
	}

	if (!altmsg || stat(altmsg, &st) == NOTOK) {
		st.st_mtime = 0;
		st.st_dev = 0;
		st.st_ino = 0;
	}
	if (pushsw) {
		push();
	}
	status = 0;
	vec[0] = mhbasename(postproc);
	closefds(3);

	for (msgnum = 0; msgnum < nfiles; msgnum++) {
		switch (sendsbr(vec, vecp, files[msgnum], &st, 1)) {
		case DONE:
			done(++status);
		case NOTOK:
			status++;  /* fall */
		case OK:
			break;
		}
	}

	context_save();
	done(status);
	return 1;
}


/*
** message sending back-end
*/
static int
sendsbr(char **vec, int vecp, char *drft, struct stat *st, int rename_drft)
{
	int status;
	char buffer[BUFSIZ];
	char *original_draft;
	char *p;  /* string pointer for building file name */

	/*
	** Save the original name of the draft file.  The name of the
	** draft file is changed to a temporary file containing the built
	** MIME message if there are attachments.  We need the original
	** name so that it can be renamed after the message is sent.
	*/
	original_draft = drft;

	/*
	** Convert the draft to a MIME message.
	** Use the mhbuild composition file for the draft if there was
	** a successful conversion because that now contains the MIME
	** message.  A nice side effect of this is that it leaves the
	** original draft file untouched so that it can be retrieved
	** and modified if desired.
	*/
	switch (attach(drft)) {
	case OK:
		drft = composition_file_name;
		break;

	case NOTOK:
		return (NOTOK);

	case DONE:
		break;
	}

	done=armed_done;
	switch (setjmp(env)) {
	case OK:
		status = sendaux(vec, vecp, drft, st) ? NOTOK : OK;
		/* rename the original draft */
		if (rename_drft && status == OK &&
				rename(original_draft, strncpy(buffer,
				m_backup(original_draft), sizeof(buffer)))
				== NOTOK) {
			advise(buffer, "unable to rename %s to", drft);
		}
		break;

	default:
		status = DONE;
		break;
	}

	done=exit;
	if (distfile) {
		unlink(distfile);
	}

	/*
	** Get rid of any temporary files that we created for attachments.
	** Also get rid of the renamed composition file that mhbuild
	** leaves as a turd.  It looks confusing, but we use the body
	** file name to help build the renamed composition file name.
	*/
	if (drft == composition_file_name) {
		clean_up_temporary_files();

		if (strlen(composition_file_name) >=
				sizeof (composition_file_name) - 6) {
			advise(NULL, "unable to remove original composition file.");
		} else {
			if (!(p = strrchr(composition_file_name, '/'))) {
				p = composition_file_name;
			} else {
				p++;
			}
			strcpy(body_file_name, p);
			*p++ = ',';
			strcpy(p, body_file_name);
			strcat(p, ".orig");

			unlink(composition_file_name);
		}
	}

	return status;
}

static int
attach(char *draft_file_name)
{
	char buf[MAXPATHLEN + 6];
	int c;
	int has_attachment;
	int has_body;
	int non_ascii; /* msg body contains non-ASCII chars */
	int length;  /* length of attachment header field name */
	char *p;

	if (!(draft_file = fopen(draft_file_name, "r"))) {
		adios(NULL, "can't open draft file `%s'.", draft_file_name);
	}

	/* We'll grow the buffer as needed. */
	field = (char *)mh_xmalloc(field_size = 256);

	/*
	** Scan the draft file for an attachment header field name.
	*/
	length = strlen(attach_hdr);
	has_attachment = 0;
	while (get_line() != EOF && *field != '\0' && *field != '-') {
		if (strncasecmp(field, attach_hdr, length)==0 &&
				field[length] == ':') {
			has_attachment = 1;
		}
	}

	/*
	** Look for at least one non-blank line in the body of the
	** message which indicates content in the body.
	** Check if body contains at least one non-blank (= not empty)
	** and if it contains any non-ASCII chars (= need MIME).
	*/
	has_body = 0;
	non_ascii = 0;
	while (get_line() != EOF) {
		for (p = field; *p != '\0'; p++) {
			if (*p != ' ' && *p != '\t') {
				has_body = 1;
			}
			if (*p > 127 || *p < 0) {
				non_ascii = 1;
			}
		}
		if (has_body && non_ascii) {
			break;  /* that's been already enough information */
		}
	}

	if (!has_attachment && non_ascii==0) {
		/* We don't need to convert it to MIME. */
		return DONE;
	}

	/*
	** Else: mimify
	*/

	/* Make names for the temporary files.  */
	strncpy(body_file_name,
			m_mktemp(toabsdir(invo_name), NULL, NULL),
			sizeof (body_file_name));
	strncpy(composition_file_name,
			m_mktemp(toabsdir(invo_name), NULL, NULL),
			sizeof (composition_file_name));

	if (has_body) {
		body_file = fopen(body_file_name, "w");
	}
	composition_file = fopen(composition_file_name, "w");

	if ((has_body && !body_file) || !composition_file) {
		clean_up_temporary_files();
		adios(NULL, "unable to open all of the temporary files.");
	}

	/* Copy non-attachment header fields to the temp composition file. */
	rewind(draft_file);
	while (get_line() != EOF && *field && *field != '-') {
		if (strncasecmp(field, attach_hdr, length) != 0 ||
				field[length] != ':') {
			fprintf(composition_file, "%s\n", field);
		}
	}
	fputs("--------\n", composition_file);

	if (has_body) {
		/* Copy the message body to the temporary file. */
		while ((c = getc(draft_file)) != EOF) {
			putc(c, body_file);
		}
		fclose(body_file);

		/* Add a mhbuild MIME composition file line for the body */
		/* charset will be discovered/guessed by buildmimeproc */
		fprintf(composition_file, "#text/plain %s\n", body_file_name);
	}

	/*
	** Now, go back to the beginning of the draft file and look for
	** header fields that specify attachments.  Add a mhbuild MIME
	** composition file for each.
	*/
	rewind(draft_file);
	while (get_line() != EOF && *field && *field != '-') {
		if (strncasecmp(field, attach_hdr, length) == 0 &&
				field[length] == ':') {
			for (p = field+length+1; *p==' ' || *p=='\t'; p++) {
				continue;
			}
			if (*p == '+') {
				/* forwarded message */
				fprintf(composition_file, "#forw [forwarded message(s)] %s\n", p);
			} else {
				/* regular attachment */
				make_mime_composition_file_entry(p);
			}
		}
	}
	fclose(composition_file);

	/*
	** We're ready to roll!  Run mhbuild on the composition file.
	** Note that mhbuild is in the context as buildmimeproc.
	*/
	sprintf(buf, "%s %s", buildmimeproc, composition_file_name);
	if (system(buf) != 0) {
		/* some problem */
		clean_up_temporary_files();
		return (NOTOK);
	}

	return (OK);
}

static void
clean_up_temporary_files(void)
{
	unlink(body_file_name);
	unlink(composition_file_name);

	return;
}

static int
get_line(void)
{
	int c;  /* current character */
	int n;  /* number of bytes in buffer */
	char *p;

	/*
	** Get a line from the input file, growing the field buffer as
	** needed.  We do this so that we can fit an entire line in the
	** buffer making it easy to do a string comparison on both the
	** field name and the field body which might be a long path name.
	*/
	for (n = 0, p = field; (c = getc(draft_file)) != EOF; *p++ = c) {
		if (c == '\n' && (c = getc(draft_file)) != ' ' && c != '\t') {
			ungetc(c, draft_file);
			c = '\n';
			break;
		}
		if (++n >= field_size - 1) {
			field = (char *)mh_xrealloc(field, field_size += 256);
			p = field + n - 1;
		}
	}
	*p = '\0';

	return (c);
}

static void
make_mime_composition_file_entry(char *file_name)
{
	FILE *fp;
	struct node *np;
	char *cp;
	char content_type[BUFSIZ];
	char cmdbuf[BUFSIZ];
	char *cmd = mimetypequeryproc;

	for (np = m_defs; np; np = np->n_next) {
		if (strcasecmp(np->n_name, mimetypequery)==0) {
			cmd = np->n_field;
			break;
		}
	}
	snprintf(cmdbuf, sizeof cmdbuf, "%s %s", cmd, file_name);

	if (!(fp = popen(cmdbuf, "r"))) {
		clean_up_temporary_files();
		adios(NULL, "unable to determine content type with `%s'",
				cmdbuf);
	}
	if (fgets(content_type, sizeof content_type, fp) &&
			(cp = strrchr(content_type, '\n'))) {
		*cp = '\0';
	} else {
		strcpy(content_type, "application/octet-stream");
		admonish(NULL, "problems with `%s', using fall back type `%s'",
				cmdbuf, content_type);
	}
	pclose(fp);

	/* TODO: don't use access(2) because it checks for ruid, not euid */
	if (access(file_name, R_OK) != 0) {
		clean_up_temporary_files();
		adios(NULL, "unable to access file `%s'", file_name);
	}

	fprintf(composition_file, "#%s; name=\"%s\" <>{attachment}",
		content_type,
		(!(cp = strrchr(file_name, '/'))) ? file_name : cp + 1);

	fprintf(composition_file, " %s\n", file_name);

	return;
}

/*
** The back-end of the message sending back-end.
** Annotate original message, and call `postproc' to send message.
*/
static int
sendaux(char **vec, int vecp, char *drft, struct stat *st)
{
	pid_t child_id;
	int status, fd;
	char backup[BUFSIZ];

	/*
	** fd collects the output of postproc, and is used for the
	** failure notice if we need to send one in alert().
	*/
	fd = pushsw ? tmp_fd() : NOTOK;

	vec[vecp++] = drft;
	if (distfile && distout(drft, distfile, backup) == NOTOK) {
		done(1);
	}
	vec[vecp] = NULL;

	switch (child_id = fork()) {
	case -1:
		/* oops -- fork error */
		adios("fork", "unable to");
		break;  /* NOT REACHED */

	case 0:
		/*
		** child process -- send it
		**
		** If fd is ok, then we are pushing and fd points to temp
		** file, so capture anything on stdout and stderr there.
		*/
		if (fd != NOTOK) {
			dup2(fd, fileno(stdout));
			dup2(fd, fileno(stderr));
			close(fd);
		}
		execvp(postproc, vec);
		fprintf(stderr, "unable to exec ");
		perror(postproc);
		_exit(-1);
		break;  /* NOT REACHED */

	default:
		/* parent process -- wait for it */
		if ((status = pidwait(child_id, NOTOK)) == OK) {
			if (annotext) {
				anno(st);
			}
		} else {
			/*
			** If postproc failed, and we have good fd (which
			** means we pushed), then mail error message
			** (and possibly the draft) back to the user.
			*/
			if (fd != NOTOK) {
				alert(drft, fd);
				close(fd);
			} else {
				advise(NULL, "message not delivered to anyone");
			}
			if (distfile) {
				unlink(drft);
				if (rename(backup, drft) == NOTOK) {
					advise(drft, "unable to rename %s to",
							backup);
				}
			}
		}
		break;
	}

	return status;
}


/*
** Mail error notification (and possibly a copy of the
** message) back to the user, using the mailproc
*/
static void
alert(char *file, int out)
{
	pid_t child_id;
	int in;
	char buf[BUFSIZ];

	switch (child_id = fork()) {
	case NOTOK:
		/* oops -- fork error */
		advise("fork", "unable to");

	case OK:
		/* child process -- send it */
		SIGNAL(SIGHUP, SIG_IGN);
		SIGNAL(SIGINT, SIG_IGN);
		SIGNAL(SIGQUIT, SIG_IGN);
		SIGNAL(SIGTERM, SIG_IGN);
		if (forwsw) {
			if ((in = open(file, O_RDONLY)) == NOTOK) {
				admonish(file, "unable to re-open");
			} else {
				lseek(out, (off_t) 0, SEEK_END);
				strncpy(buf, "\nMessage not delivered to anyone.\n", sizeof(buf));
				write(out, buf, strlen(buf));
				strncpy(buf, "\n------- Unsent Draft\n\n", sizeof(buf));
				write(out, buf, strlen(buf));
				cpydgst(in, out, file, "temporary file");
				close(in);
				strncpy(buf, "\n------- End of Unsent Draft\n", sizeof(buf));
				write(out, buf, strlen(buf));
				if (rename(file, strncpy(buf, m_backup(file), sizeof(buf))) == NOTOK) {
					admonish(buf, "unable to rename %s to", file);
				}
			}
		}
		lseek(out, (off_t) 0, SEEK_SET);
		dup2(out, fileno(stdin));
		close(out);
		/* create subject for error notification */
		snprintf(buf, sizeof(buf), "send failed on %s",
				forwsw ? "enclosed draft" : file);
		execlp(mailproc, mhbasename(mailproc), getusername(),
				"-subject", buf, NULL);
		fprintf(stderr, "unable to exec ");
		perror(mailproc);
		_exit(-1);

	default:  /* no waiting... */
		break;
	}
}


static int
tmp_fd(void)
{
	int fd;
	char *tfile = NULL;

	tfile = m_mktemp2(NULL, invo_name, &fd, NULL);
	if (tfile == NULL) return NOTOK;
	fchmod(fd, 0600);

	if (debugsw) {
		advise(NULL, "temporary file %s selected", tfile);
	} else {
		if (unlink(tfile) == NOTOK) {
			advise(tfile, "unable to remove");
		}
	}

	return fd;
}


static void
anno(struct stat *st)
{
	struct stat st2;
	char *msgs, *folder;
	char buf[BUFSIZ];

	if (altmsg && (stat(altmsg, &st2) == NOTOK ||
			st->st_mtime != st2.st_mtime ||
			st->st_dev != st2.st_dev ||
			st->st_ino != st2.st_ino)) {
		if (debugsw) {
			admonish(NULL, "$mhaltmsg mismatch");
		}
		return;
	}

	if (!(folder = getenv("mhfolder")) || !*folder) {
		if (debugsw) {
			admonish(NULL, "$mhfolder not set");
		}
		return;
	}
	if (!(msgs = getenv("mhmessages")) || !*msgs) {
		if (debugsw) {
			admonish(NULL, "$mhmessages not set");
		}
		return;
	}
	if (debugsw) {
		advise(NULL, "annotate as `%s': %s %s", annotext,
				folder, msgs);
	}
	snprintf(buf, sizeof buf, "anno -comp '%s' '+%s' %s",
			annotext, folder, msgs);
	if (system(buf) != 0) {
		advise(NULL, "unable to annotate");
	}
}


static void
armed_done(int status)
{
	longjmp(env, status ? status : NOTOK);

	exit(status);
}
